import abc
import math
import typing


class Converter(abc.ABC):
    @abc.abstractproperty
    def REGEX(self) -> str:
        ...

    @abc.abstractmethod
    def convert(self, string: str):
        ...


class StringConverter(Converter):
    """Converter for string URL parameters.

    Parameters
    ----------
        length : Optional :class:`int`
            Required length of the string.
            Default: ``None``

        allow_slash : Optional :class:`bool`
            Allow slashes in the string.
            Default: ``False``

    Attributes
    ----------
        length : Optional :class:`int`
            Required length of the string.

        allow_slash : :class:`bool`
            Allow slashes in the string.

        REGEX : :class:`str`
            Regex for the route :meth:`~baguette.router.Route.build_regex`.

    """

    REGEX = r"[^\/]+"

    def __init__(
        self,
        length: typing.Optional[int] = None,
        allow_slash: bool = False,
    ):
        self.length = length
        self.allow_slash = allow_slash
        if self.allow_slash:
            self.REGEX = r".+"

    def convert(self, string: str):
        """Converts the string of the URL parameter and validates the value.

        Parameters
        ----------
            string : :class:`str`
                URL parameter to convert.

        Returns
        -------
            :class:`str`
                Converted URL parameter.

        Raises
        ------
            ValueError
                :attr:`length` is specified and the URL parameter has a
                different length than :attr:`length`.

            ValueError
                :attr:`allow_slash` is ``False`` and the URL parameter
                contains slashes.

        """

        if self.length is not None and len(string) != self.length:
            raise ValueError(
                f"Expected string of length {self.length}. Got {len(string)}"
            )

        if not self.allow_slash and "/" in string:
            raise ValueError(f"Expected string without '/'. Got {string!r}")

        return str(string)


class PathConverter(Converter):
    """Converter for string URL parameters.

    .. note::
        This is approximatly equivalent to a
        :class:`~baguette.converters.StringConverter` with
        :attr:`~baguette.converters.StringConverter.allow_slash` set to ``True``

    Attributes
    ----------
        REGEX : :class:`str`
            Regex for the route :meth:`~baguette.router.Route.build_regex`.

    """

    REGEX = r".+"

    def __init__(self):
        pass

    def convert(self, string: str):
        """Converts the string of the URL parameter and validates the value.

        Parameters
        ----------
            string : :class:`str`
                URL parameter to convert.

        Returns
        -------
            :class:`str`
                Converted URL parameter.

        """

        return string.strip("/")


class IntegerConverter(Converter):
    """Converter for integer URL parameters.

    Parameters
    ----------
        signed : Optional :class:`bool`
            Whether to accept integers starting with ``+`` or ``-``.
            Default: ``False``

        min : Optional :class:`int`
            Minimum value of the integer.
            Default: ``None``

        max : Optional :class:`int`
            Maximum value of the integer.
            Default: ``None``

    Attributes
    ----------
        signed : :class:`bool`
            Whether to accept integers starting with ``+`` or ``-``.

        min : Optional :class:`int`
            Minimum value of the integer.

        max : Optional :class:`int`
            Maximum value of the integer.

        REGEX : :class:`str`
            Regex for the route :meth:`~baguette.router.Route.build_regex`.

    """

    REGEX = r"[\+-]?\d+"

    def __init__(
        self,
        signed: bool = False,
        min: typing.Optional[int] = None,
        max: typing.Optional[int] = None,
    ):
        self.signed = signed
        self.min = min
        self.max = max

    def convert(self, string: str):
        """Converts the string of the URL parameter and validates the value.

        Parameters
        ----------
            string : :class:`str`
                URL parameter to convert.

        Returns
        -------
            :class:`int`
                Converted URL parameter.

        Raises
        ------
            ValueError
                :attr:`signed` is ``False`` and the URL parameter starts
                with ``+`` or ``-``.

            ValueError
                Couldn't convert the URL parameter to an integer.

            ValueError
                :attr:`min` is specified and the URL parameter
                is lower then :attr:`min`.

            ValueError
                :attr:`max` is specified and the URL parameter
                is higher then :attr:`max`.

        """

        if not self.signed and (string.strip()[0] in "+-"):
            raise ValueError(
                "Expected unsigned integer. Got integer starting with "
                + string.strip()[0]
            )

        integer = int(string)

        if self.min is not None and integer < self.min:
            raise ValueError(f"Expected integer higher than {self.min}")

        if self.max is not None and integer > self.max:
            raise ValueError(f"Expected integer lower than {self.max}")

        return integer


class FloatConverter(Converter):
    """Converter for float URL parameters.

    Parameters
    ----------
        signed : Optional :class:`bool`
            Whether to accept floats starting with ``+`` or ``-``.
            Default: ``False``

        min : Optional :class:`float`
            Minimum value of the float.
            Default: ``None``

        max : Optional :class:`float`
            Maximum value of the float.
            Default: ``None``

        allow_infinity : Optional :class:`bool`
            Whether to accept floats that are ``inf`` or ``-inf``.
            Default: ``False``

        allow_nan : Optional :class:`bool`
            Whether to accept floats that are ``NaN``.
            Default: ``False``

    Attributes
    ----------
        signed : :class:`bool`
            Whether to accept floats starting with ``+`` or ``-``.

        min : Optional :class:`float`
            Minimum value of the float.

        max : Optional :class:`float`
            Maximum value of the float.

        allow_infinity : :class:`bool`
            Whether to accept floats that are ``inf`` or ``-inf``.

        allow_nan : :class:`bool`
            Whether to accept floats that are ``NaN``.

        REGEX : :class:`str`
            Regex for the route :meth:`~baguette.router.Route.build_regex`.

    """

    REGEX = r"[\+-]?\d*\.?\d*"

    def __init__(
        self,
        signed: bool = False,
        min: typing.Optional[float] = None,
        max: typing.Optional[float] = None,
        allow_infinity: bool = False,
        allow_nan: bool = False,
    ):
        self.signed = signed
        self.min = min
        self.max = max
        self.allow_infinity = allow_infinity
        self.allow_nan = allow_nan

    def convert(self, string: str):
        """Converts the string of the URL parameter and validates the value.

        Parameters
        ----------
            string : :class:`str`
                URL parameter to convert.

        Returns
        -------
            :class:`float`
                Converted URL parameter.

        Raises
        ------
            ValueError
                :attr:`signed` is ``False`` and the URL parameter starts
                with ``+`` or ``-``.

            ValueError
                Couldn't convert the URL parameter to an float.

            ValueError
                :attr:`min` is specified and the URL parameter
                is lower then :attr:`min`.

            ValueError
                :attr:`max` is specified and the URL parameter
                is higher then :attr:`max`.

            ValueError
                :attr:`allow_infinity` is ``False`` and the URL parameter
                is ``inf`` or ``-inf``.

            ValueError
                :attr:`allow_nan` is ``False`` and the URL parameter
                is ``nan``.

        """

        if not self.signed and (string.strip()[0] in "+-"):
            raise ValueError(
                "Expected unsigned float. Got float starting with "
                + string.strip()[0]
            )

        number = float(string)

        if self.min is not None and number < self.min:
            raise ValueError(f"Expected float higher than {self.min}")

        if self.max is not None and number > self.max:
            raise ValueError(f"Expected float lower than {self.max}")

        if not self.allow_infinity and math.isinf(number):
            raise ValueError(f"Expected a non-infinity value. Got {number}")

        if not self.allow_nan and math.isnan(number):
            raise ValueError(f"Expected a non-NaN value. Got {number}")

        return number



import inspect
import ssl
import traceback
import typing

from .headers import Headers, make_headers
from .httpexceptions import BadRequest, HTTPException, InternalServerError
from .request import Request
from .responses import (
    FileResponse,
    Response,
    make_error_response,
    make_response,
)
from .router import Route, Router
from .types import Handler, HeadersType, Receive, Result, Scope, Send
from .view import View


class Baguette:
    """Implements an ASGI application.

    This class is the main class for any application written with the baguette
    framework

    Parameters
    ----------
        debug: :class:`bool`
            Whether to run the application in debug mode.
            Default: ``False``.

        default_headers: :class:`list` of ``(str, str)`` tuples, \
        :class:`dict` or :class:`Headers`
            Default headers to include in every request.
            Default: No headers.

        static_url_path: :class:`str`
            URL path for the static file handler.
            Default: ``"static"``.

        static_directory: :class:`str`
            Path to the folder containing static files.
            Default: ``"static"``.

        error_response_type: :class:`str`
            Type of response to use in case of error.
            One of: ``"plain"``, ``"json"``, ``"html"``.
            Default: ``"plain"``.

        error_include_description: :class:`bool`
            Whether to include the error description in the response
            in case of error.
            If debug is ``True``, this will also be ``True``.
            Default: ``True``.


    Attributes
    ----------
        router: :class:`~baguette.router.Router`
            The URL router of the app.

        debug: :class:`bool`
            Whether the application is running in debug mode.

        default_headers: :class:`Headers`
            Default headers included in every response.

        static_url_path: :class:`str`
            URL path for the static file handler.

        static_directory: :class:`str`
            Path to the folder containing static files.

        error_response_type: :class:`str`
            Type of response to use in case of error.
            One of: ``"plain"``, ``"json"``, ``"html"``

        error_include_description: :class:`bool`
            Whether the error description is included in the response
            in case of error.
            If debug is ``True``, this will also be ``True``.
    """

    def __init__(
        self,
        *,
        debug: bool = False,
        default_headers: HeadersType = None,
        static_url_path: str = "static",
        static_directory: str = "static",
        error_response_type: str = "plain",
        error_include_description: bool = True,
    ):
        self.router = Router()
        self.debug = debug
        self.default_headers: Headers = make_headers(default_headers)

        self.static_url_path = static_url_path
        self.static_directory = static_directory

        self.add_route(
            path=f"{self.static_url_path}/<filename:path>",
            handler=self.handle_static_file,
            name="static",
        )

        if error_response_type not in ("plain", "json", "html"):
            raise ValueError(
                "Bad response type. Must be one of: 'plain', 'json', 'html'"
            )

        self.error_response_type = error_response_type
        self.error_include_description = error_include_description or self.debug

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        """Entry point of the ASGI application."""

        asgi_handlers = {
            "http": self._handle_http,
            "lifespan": self._handle_lifespan,
        }

        asgi_handler = asgi_handlers.get(scope["type"])
        if asgi_handler is None:
            raise NotImplementedError(
                "{0!r} scope is not supported".format(scope["type"])
            )

        await asgi_handler(scope, receive, send)

    async def _handle_http(self, scope: Scope, receive: Receive, send: Send):
        """Handles rquests where ``scope["type"] == "http"``."""

        request = Request(self, scope, receive)
        response = await self.handle_request(request)
        await response.send(send)

    async def _handle_lifespan(
        self, scope: Scope, receive: Receive, send: Send
    ):
        """Handles rquests where ``scope["type"] == "lifespan"``."""

        while True:
            message = await receive()
            if message["type"] == "lifespan.startup":
                try:
                    await self.startup()
                except Exception as e:
                    await send(
                        {
                            "type": "lifespan.startup.failed",
                            "message": str(e),
                        }
                    )
                else:
                    await send({"type": "lifespan.startup.complete"})

            elif message["type"] == "lifespan.shutdown":
                try:
                    await self.shutdown()
                except Exception as e:
                    await send(
                        {
                            "type": "lifespan.shutdown.failed",
                            "message": str(e),
                        }
                    )
                else:
                    await send({"type": "lifespan.shutdown.complete"})

                return

    async def startup(self):
        """Runs on application startup.

        This will be executed when you start the application.
        For example, you can connect to a database.

        .. versionadded:: 0.1.0
        """

    async def shutdown(self):
        """Runs on application shutdown.

        This will be executed when you stop the application.
        For example, you can disconnect from a database.


        .. versionadded:: 0.1.0
        """

    async def handle_request(self, request: Request) -> Response:
        """Handles a request and returns a response.

        Parameters
        ----------
            request: :class:`Request`
                The request to handle.

        Returns
        -------
            :class:`Response`
                A response.
        """

        result = await self.dispatch(request)
        response = make_response(result)
        response.headers += self.default_headers
        return response

    async def dispatch(self, request: Request) -> Result:
        """Dispatches a request to the correct handler and return its result.

        Parameters
        ----------
            request: :class:`Request`
                The request to handle.

        Returns
        -------
            Anything described in :ref:`responses`
                The handler function return value.
        """

        try:
            route: Route = self.router.get(request.path, request.method)
            handler: Handler = route.handler

            try:
                kwargs = route.convert(request.path)
            except ValueError:
                raise BadRequest(description="Failed to convert URL parameters")
            kwargs["request"] = request
            if not route.handler_is_class:
                kwargs = {
                    k: v for k, v in kwargs.items() if k in route.handler_kwargs
                }

            return await handler(**kwargs)

        except HTTPException as http_exception:
            return make_error_response(
                http_exception,
                type_=self.error_response_type,
                include_description=self.error_include_description,
                traceback="".join(
                    traceback.format_tb(http_exception.__traceback__)
                )
                if self.debug and http_exception.status_code >= 500
                else None,
            )
        except Exception as exception:
            traceback.print_exc()
            http_exception = InternalServerError()
            return make_error_response(
                http_exception,
                type_=self.error_response_type,
                include_description=self.error_include_description,
                traceback="".join(traceback.format_tb(exception.__traceback__))
                if self.debug
                else None,
            )

    def add_route(
        self,
        path: str,
        handler: Handler,
        methods: typing.List[str] = None,
        name: str = None,
        defaults: dict = None,
    ) -> Route:
        """Adds a route to the application router.

        Parameters
        ----------
            handler: Async callable
                An asynchronous callable (function or class)
                that can handle a request.

            path: :class:`str`
                The path that the handler will handle.
                Can be dynamic, see :ref:`dynamic_routing`.

            methods: :class:`list` of :class:`str`
                Allowed methods for this path.
                Default: ``["GET", "HEAD"]``.

            name: :class:`str`
                Name of the route.
                Default: handler function name.

            defaults: Optional :class:`dict`
                Default arguments to give to your handler.
                Default: ``{}``.

        Returns
        -------
            :class:`~baguette.router.Route`
                The created route.
        """

        return self.router.add_route(
            path=path,
            name=name,
            handler=handler,
            methods=methods or ["GET", "HEAD"],
            defaults=defaults or {},
        )

    def route(
        self,
        path: str,
        methods: typing.List[str] = None,
        name: str = None,
        defaults: dict = None,
    ):
        """Adds the handler function to the router with the given path.

        Parameters
        ----------
            path: :class:`str`
                The path that the handler will handle.
                Can be dynamic, see :ref:`dynamic_routing`.

            methods: Optional :class:`list` of :class:`str`
                Allowed methods for this path.
                Default: ``["GET", "HEAD"]``.

            name: Optional :class:`str`
                Name of the route.
                Default: handler function name.

            defaults: Optional :class:`dict`
                Default arguments to give to your handler.
                Default: ``{}``.

        .. versionchanged:: 0.0.3
            Renamed from ``Baguette.endpoint`` to :meth:`Baguette.route`

        """

        def decorator(func_or_class):
            if inspect.isclass(func_or_class) and issubclass(
                func_or_class, View
            ):
                handler: Handler = func_or_class(self)
                allowed_methods = handler.methods

            else:
                allowed_methods = methods
                handler: Handler = func_or_class

            self.add_route(
                path=path,
                name=name or func_or_class.__name__,
                handler=handler,
                methods=allowed_methods,
                defaults=defaults or {},
            )

            return func_or_class

        return decorator

    async def handle_static_file(self, filename: str) -> FileResponse:
        return FileResponse(self.static_directory, filename)

    def run(
        self,
        *,
        host: str = "127.0.0.1",
        port: int = 8000,
        uds: str = None,
        fd: int = None,
        debug: bool = None,
        headers=None,
        loop: str = "auto",
        http: str = "auto",
        ws: str = "auto",
        env_file: str = None,
        log_config=None,
        log_level: str = None,
        access_log: bool = True,
        use_colors: bool = None,
        workers: int = None,
        proxy_headers: bool = True,
        forwarded_allow_ips: str = None,
        root_path: str = "",
        limit_concurrency: int = None,
        limit_max_requests: int = None,
        backlog: int = 2048,
        timeout_keep_alive: int = 5,
        timeout_notify: int = 30,
        callback_notify: typing.Callable = None,
        ssl_keyfile: str = None,
        ssl_certfile: str = None,
        ssl_keyfile_password: str = None,
        ssl_version: int = None,
        ssl_cert_reqs: int = ssl.CERT_NONE,
        ssl_ca_certs: str = None,
        ssl_ciphers: str = "TLSv1",
    ):
        """Runs the server with uvicorn.

        .. warning::
            You need ``uvicorn`` installed in order to use this.
            See :ref:`installing_asgi_server`.

        Keyword Arguments
        -----------------
            host: Optional :class:`str`
                Bind socket to this host.
                IPv6 addresses are supported.
                Default: ``127.0.0.1``

            port: Optional :class:`int`
                Bind to a socket with this port.
                Default: ``8000``.

            uds: Optional :class:`str`
                Bind to a UNIX domain socket.
                Default: ``None``.

            fd: Optional :class:`int`
                Bind to socket from this file descriptor.
                Default: ``None``.

            debug: Optional :class:`bool`
                Update the app.debug variable while the app is running.
                Default: ``None``.

            headers: Optional :class:`list` of ``(str, str)`` tuples, \
            :class:`dict` or :class:`Headers`
                Add headers to every response while the app is running.
                Default: ``None``.

            loop: Optional ``"auto"``, ``"asyncio"`` or ``"uvloop"``
                Event loop implementation. The uvloop implementation provides
                greater performance, but is not compatible with Windows or PyPy.
                Default: ``"auto"``

            http: Optional ``"auto"``, ``"h11"`` or ``"httptools"``
                HTTP protocol implementation. The httptools implementation
                provides greater performance, but it not compatible with PyPy,
                and requires compilation on Windows.
                Default: ``"auto"``

            ws: Optional ``"auto"``, ``"none"``, ``"websockets"`` \
            or ``"wsproto"``
                WebSocket protocol implementation. Either of the ``websockets``
                and ``wsproto`` packages are supported.
                Use ``"none"`` to deny all websocket requests.
                Default: ``"auto"``

            env_file: Optional :class:`str` (path to file)
                Environment configuration file. Loaded with
                `python-dotenv <https://pypi.org/project/python-dotenv/>`_
                Default: ``None``

            log_config: Optional :func:`~logging.config.dictConfig` formats, \
            ``.json`` and ``.yaml`` file paths or \
            :func:`~logging.config.fileConfig` formats.
                Logging configuration.
                Default: `uvicorn.config.LOGGING_CONFIG \
                <https://github.com/encode/uvicorn/blob/master/uvicorn/config.py#L68>`_

            log_level: ``"critical"``, ``"error"``, ``"warning"``, ``"info"``, \
            ``"debug"`` or ``"trace"``
                Application log level.
                Default: ``"info"``

            access_log: Optional :class:`bool`
                Whether to log every request.
                Default: ``True``

            use_colors: Optional :class:`bool`
                Whether to enable colorized formatting of the log records,
                in case this is not set it will be auto-detected.
                This option is ignored if  ``log_config`` is given.
                Default: ``None``

            workers: Optional :class:`int`
                Number of worker processes to use.
                Default: ``WEB_CONCURRENCY`` environment variable
                if available, or ``1``

            proxy_headers: Optional :class:`bool`
                Whether to enable ``X-Forwarded-Proto``, ``X-Forwarded-For`` and
                ``X-Forwarded-Port`` headers to populate remote address info.
                Restricted to only trusting connecting IPs in the
                ``forwarded_allow_ips`` parameter.
                Default: ``True``

            forwarded_allow_ips: Optional :class:`str`
                Comma separated list of IPs to trust with proxy headers.
                A wildcard ``"*"`` means always trust.
                Default: ``FORWARDED_ALLOW_IPS`` environment variable
                if available, or ``127.0.0.1``

            ssl_keyfile: Optional :class:`str` (path to file)
                SSL key file
                Default: ``None``

            ssl_certfile: Optional :class:`str` (path to file)
                SSL certificate file
                Default: ``None``

            ssl_keyfile_password: Optional :class:`str`
                Password to decrypt the ssl key
                Default: ``None``

            ssl_version: Optional :class:`int`
                SSL version to use. See :mod:`ssl` module.
                Default: :data:`ssl.PROTOCOL_TLS`

            ssl_cert_reqs: Optional :class:`int`
                Whether client certificate is required.
                One of :class:`ssl.VerifyMode` values,
                a constant of the :mod:`ssl` module that starts with ``CERT_``.
                Default: :data:`ssl.CERT_NONE`

            ssl_ca_certs: Optional :class:`str` (path to file)
                CA certificates file
                Default: ``None``

            ssl_ciphers: Optional :class:`str`
                Ciphers to use. See :mod:`ssl` module.
                Default: ``"TLSv1"``

        .. note::
            Doesn't support reloading, if you want reloading,
            run the application from the console with uvicorn
            and add the ``--reload`` flag.

        .. versionadded:: 0.1.2

        """

        try:
            import uvicorn
        except ModuleNotFoundError:  # pragma: no cover
            raise RuntimeError("Install uvicorn to use app.run()")

        last_debug = self.debug
        if isinstance(debug, bool):
            self.debug = debug

        last_default_headers = self.default_headers
        headers = make_headers(self.default_headers) + make_headers(headers)
        self.default_headers = headers

        log_config = log_config or uvicorn.config.LOGGING_CONFIG
        workers = workers or 1
        ssl_version = ssl_version or uvicorn.config.SSL_PROTOCOL_VERSION

        uvicorn.run(
            self,
            host=host,
            port=port,
            uds=uds,
            fd=fd,
            loop=loop,
            http=http,
            ws=ws,
            env_file=env_file,
            log_config=log_config,
            log_level=log_level,
            access_log=access_log,
            use_colors=use_colors,
            workers=workers,
            proxy_headers=proxy_headers,
            forwarded_allow_ips=forwarded_allow_ips,
            root_path=root_path,
            limit_concurrency=limit_concurrency,
            limit_max_requests=limit_max_requests,
            backlog=backlog,
            timeout_keep_alive=timeout_keep_alive,
            timeout_notify=timeout_notify,
            callback_notify=callback_notify,
            ssl_keyfile=ssl_keyfile,
            ssl_certfile=ssl_certfile,
            ssl_keyfile_password=ssl_keyfile_password,
            ssl_version=ssl_version,
            ssl_cert_reqs=ssl_cert_reqs,
            ssl_ca_certs=ssl_ca_certs,
            ssl_ciphers=ssl_ciphers,
        )

        self.debug = last_debug
        self.default_headers = last_default_headers



"""
Exceptions for HTTP status code over 400 (4xx client errors and 5xx server errors) from the `http module <https://docs.python.org/3/library/http.html>`_.

=========== =============================== ==================================== ============================================================================================================================================================================================================== ===============================================================================================
Status code               Name                       Exception class                                                                                                           Description                                                                                                                                               RFC Link
=========== =============================== ==================================== ============================================================================================================================================================================================================== ===============================================================================================
400         Bad Request                     :exc:`BadRequest`                    Bad request syntax or unsupported method.                                                                                                                                                                      `RFC 7231, Section 6.5.1   <https://tools.ietf.org/html/rfc7231.html#section-6.5.1>`_
401         Unauthorized                    :exc:`Unauthorized`                  No permission -- see authorization schemes.                                                                                                                                                                    `RFC 7235 , Section 3.1    <https://tools.ietf.org/html/rfc7235.html#section-3.1>`_
402         Payment Required                :exc:`PaymentRequired`               No payment -- see charging schemes.                                                                                                                                                                            `RFC 7231, Section 6.5.2   <https://tools.ietf.org/html/rfc7231.html#section-6.5.2>`_
403         Forbidden                       :exc:`Forbidden`                     Request forbidden -- authorization will not help.                                                                                                                                                              `RFC 7231, Section 6.5.3   <https://tools.ietf.org/html/rfc7231.html#section-6.5.3>`_
404         Not Found                       :exc:`NotFound`                      Nothing matches the given URI.                                                                                                                                                                                 `RFC 7231, Section 6.5.4   <https://tools.ietf.org/html/rfc7231.html#section-6.5.4>`_
405         Method Not Allowed              :exc:`MethodNotAllowed`              Specified method is invalid for this resource.                                                                                                                                                                 `RFC 7231, Section 6.5.5   <https://tools.ietf.org/html/rfc7231.html#section-6.5.5>`_
406         Not Acceptable                  :exc:`NotAcceptable`                 URI not available in preferred format.                                                                                                                                                                         `RFC 7231, Section 6.5.6   <https://tools.ietf.org/html/rfc7231.html#section-6.5.6>`_
407         Proxy Authentication Required   :exc:`ProxyAuthenticationRequired`   You must authenticate with this proxy before proceeding.                                                                                                                                                       `RFC 7235 , Section 3.2    <https://tools.ietf.org/html/rfc7235.html#section-3.2>`_
408         Request Timeout                 :exc:`RequestTimeout`                Request timed out; try again later.                                                                                                                                                                            `RFC 7231, Section 6.5.7   <https://tools.ietf.org/html/rfc7231.html#section-6.5.7>`_
409         Conflict                        :exc:`Conflict`                      Request conflict.                                                                                                                                                                                              `RFC 7231, Section 6.5.8   <https://tools.ietf.org/html/rfc7231.html#section-6.5.8>`_
410         Gone                            :exc:`Gone`                          URI no longer exists and has been permanently removed.                                                                                                                                                         `RFC 7231, Section 6.5.9   <https://tools.ietf.org/html/rfc7231.html#section-6.5.9>`_
411         Length Required                 :exc:`LengthRequired`                Client must specify Content-Length header.                                                                                                                                                                     `RFC 7231, Section 6.5.10  <https://tools.ietf.org/html/rfc7231.html#section-6.5.10>`_
412         Precondition Failed             :exc:`PreconditionFailed`            Precondition in headers is false.                                                                                                                                                                              `RFC 7232, Section 4.2     <https://tools.ietf.org/html/rfc7232.html#section-4.2>`_
413         Request Entity Too Large        :exc:`RequestEntityTooLarge`         Entity is too large.                                                                                                                                                                                           `RFC 7231, Section 6.5.11  <https://tools.ietf.org/html/rfc7231.html#section-6.5.11>`_
414         Request-URI Too Long            :exc:`RequestURITooLong`             URI is too long.                                                                                                                                                                                               `RFC 7231, Section 6.5.12  <https://tools.ietf.org/html/rfc7231.html#section-6.5.12>`_
415         Unsupported Media Type          :exc:`UnsupportedMediaType`          Entity body in unsupported format.                                                                                                                                                                             `RFC 7231, Section 6.5.13  <https://tools.ietf.org/html/rfc7231.html#section-6.5.13>`_
416         Requested Range Not Satisfiable :exc:`RequestedRangeNotSatisfiable`  Cannot satisfy request range.                                                                                                                                                                                  `RFC 7233, Section 4.4     <https://tools.ietf.org/html/rfc7233.html#section-4.4>`_
417         Expectation Failed              :exc:`ExpectationFailed`             Expect condition could not be satisfied.                                                                                                                                                                       `RFC 7231, Section 6.5.14  <https://tools.ietf.org/html/rfc7231.html#section-6.5.14>`_
418         I'm a Teapot                    :exc:`IMATeapot`                     Server refuses to brew coffee because it is a teapot (1998 April Fools').                                                                                                                                      `RFC 2324, Section 2.3.2   <https://tools.ietf.org/html/rfc2324.html#section-2.3.2>`_
421         Misdirected Request             :exc:`MisdirectedRequest`            Server is not able to produce a response.                                                                                                                                                                      `RFC 7540, Section 9.1.2   <https://tools.ietf.org/html/rfc7540.html#section-9.1.2>`_
422         Unprocessable Entity            :exc:`UnprocessableEntity`           Server understands the content type of the request entity and the syntax of the request entity is correct but was unable to process the contained instructions.                                                `RFC 4918, Section 11.2    <https://tools.ietf.org/html/rfc4918.html#section-11.2>`_
423         Locked                          :exc:`Locked`                        Source or destination resource of a method is locked.                                                                                                                                                          `RFC 4918, Section 11.3    <https://tools.ietf.org/html/rfc4918.html#section-11.3>`_
424         Failed Dependency               :exc:`FailedDependency`              Request method could not be performed on the resource because the requested action depended on another action that failed.                                                                                     `RFC 4918, Section 11.4    <https://tools.ietf.org/html/rfc4918.html#section-11.4>`_
425         Too Early                       :exc:`TooEarly`                      Server is unwilling to risk processing a request that might be replayed.                                                                                                                                       `RFC 8470, Section 5.2     <https://tools.ietf.org/html/rfc8470.html#section-5.2>`_
426         Upgrade Required                :exc:`UpgradeRequired`               Server refuses to perform the request using the current protocol but might be willing to do so after the client upgrades to a different protocol.                                                              `RFC 7231, Section 6.5.15  <https://tools.ietf.org/html/rfc7231.html#section-6.5.15>`_
428         Precondition Required           :exc:`PreconditionRequired`          Origin server requires the request to be conditional.                                                                                                                                                          `RFC 6585, Section 3       <https://tools.ietf.org/html/rfc6585.html#section-3>`_
429         Too Many Requests               :exc:`TooManyRequests`               User has sent too many requests in a given amount of time ("rate limiting").                                                                                                                                   `RFC 6585, Section 4       <https://tools.ietf.org/html/rfc6585.html#section-4>`_
431         Request Header Fields Too Large :exc:`RequestHeaderFieldsTooLarge`   Server is unwilling to process the request because its header fields are too large.                                                                                                                            `RFC 6585, Section 5       <https://tools.ietf.org/html/rfc6585.html#section-5>`_
451         Unavailable For Legal Reasons   :exc:`UnavailableForLegalReasons`    Server is denying access to the resource as a consequence of a legal demand.                                                                                                                                   `RFC 7725, Section 3       <https://tools.ietf.org/html/rfc7725.html#section-3>`_
500         Internal Server Error           :exc:`InternalServerError`           Server got itself in trouble.                                                                                                                                                                                  `RFC 7231, Section 6.6.1   <https://tools.ietf.org/html/rfc7231.html#section-6.6.1>`_
501         Not Implemented                 :exc:`NotImplemented`                Server does not support this operation.                                                                                                                                                                        `RFC 7231, Section 6.6.2   <https://tools.ietf.org/html/rfc7231.html#section-6.6.2>`_
502         Bad Gateway                     :exc:`BadGateway`                    Invalid responses from another server/proxy.                                                                                                                                                                   `RFC 7231, Section 6.6.3   <https://tools.ietf.org/html/rfc7231.html#section-6.6.3>`_
503         Service Unavailable             :exc:`ServiceUnavailable`            Server cannot process the request due to a high load.                                                                                                                                                          `RFC 7231, Section 6.6.4   <https://tools.ietf.org/html/rfc7231.html#section-6.6.4>`_
504         Gateway Timeout                 :exc:`GatewayTimeout`                Gateway server did not receive a timely response.                                                                                                                                                              `RFC 7231, Section 6.6.5   <https://tools.ietf.org/html/rfc7231.html#section-6.6.5>`_
505         HTTP Version Not Supported      :exc:`HTTPVersionNotSupported`       Cannot fulfill request.                                                                                                                                                                                        `RFC 7231, Section 6.6.6   <https://tools.ietf.org/html/rfc7231.html#section-6.6.6>`_
506         Variant Also Negotiates         :exc:`VariantAlsoNegotiates`         Server has an internal configuration error: the chosen variant resource is configured to engage in transparent content negotiation itself, and is therefore not a proper end point in the negotiation process. `RFC 2295, Section 8.1     <https://tools.ietf.org/html/rfc2295.html#section-8.1>`_
507         Insufficient Storage            :exc:`InsufficientStorage`           Method could not be performed on the resource because the server is unable to store the representation needed to successfully complete the request.                                                            `RFC 4918, Section 11.5    <https://tools.ietf.org/html/rfc4918.html#section-11.5>`_
508         Loop Detected                   :exc:`LoopDetected`                  Server terminated an operation because it encountered an infinite loop while processing a request.                                                                                                             `RFC 5842, Section 7.2     <https://tools.ietf.org/html/rfc5842.html#section-7.2>`_
510         Not Extended                    :exc:`NotExtended`                   Server terminated an operation because it encountered an infinite loop while processing a request.                                                                                                             `RFC 2774, Section 7       <https://tools.ietf.org/html/rfc2774.html#section-7>`_
511         Network Authentication Required :exc:`NetworkAuthenticationRequired` Client needs to authenticate to gain network access.                                                                                                                                                           `RFC 6585, Section 6       <https://tools.ietf.org/html/rfc6585.html#section-6>`_
=========== =============================== ==================================== ============================================================================================================================================================================================================== ===============================================================================================
"""


import http


class HTTPException(Exception):
    """Base class for HTTP exceptions.


    Attributes
    ----------
        status_code: :class:`int`
            HTTP status code.

        name: :class:`str`
            Name of the HTTP exception.

        description: :class:`str`
            Description of the HTTP exception.
    """

    def __init__(
        self, status_code: int, name: str = None, description: str = None
    ):
        if name is None:
            name = http.HTTPStatus(status_code).phrase
        if description is None:
            description = http.HTTPStatus(status_code).description
        self.status_code = status_code
        self.name = name
        self.description = description

    def __repr__(self) -> str:
        return (
            "HTTP Exception: {0.status_code} {0.name}{1}{0.description}".format(
                self, ": " if self.description else ""
            )
        )


class BadRequest(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(400, name=name, description=description)


class Unauthorized(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(401, name=name, description=description)


class PaymentRequired(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(402, name=name, description=description)


class Forbidden(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(403, name=name, description=description)


class NotFound(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(404, name=name, description=description)


class MethodNotAllowed(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(405, name=name, description=description)


class NotAcceptable(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(406, name=name, description=description)


class ProxyAuthenticationRequired(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(407, name=name, description=description)


class RequestTimeout(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(408, name=name, description=description)


class Conflict(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(409, name=name, description=description)


class Gone(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(410, name=name, description=description)


class LengthRequired(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(411, name=name, description=description)


class PreconditionFailed(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(412, name=name, description=description)


class RequestEntityTooLarge(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(413, name=name, description=description)


class RequestURITooLong(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(414, name=name, description=description)


class UnsupportedMediaType(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(415, name=name, description=description)


class RequestedRangeNotSatisfiable(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(416, name=name, description=description)


class ExpectationFailed(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(417, name=name, description=description)


class IMATeapot(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        if name is None:
            name = "I'm a Teapot"  # For py < 3.9
        if description is None:
            description = (
                "Server refuses to brew coffee because it is a teapot."
            )
        super().__init__(418, name=name, description=description)


class MisdirectedRequest(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        if name is None:
            name = "Misdirected Request"  # For py < 3.7
        if description is None:
            description = "Server is not able to produce a response"
        super().__init__(421, name=name, description=description)


class UnprocessableEntity(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(422, name=name, description=description)


class Locked(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(423, name=name, description=description)


class FailedDependency(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(424, name=name, description=description)


class TooEarly(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        if name is None:
            name = "Too Early"  # For py < 3.9
        if description is None:
            description = "Server is unwilling to risk processing a request that might be replayed"
        super().__init__(425, name=name, description=description)


class UpgradeRequired(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(426, name=name, description=description)


class PreconditionRequired(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(428, name=name, description=description)


class TooManyRequests(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(429, name=name, description=description)


class RequestHeaderFieldsTooLarge(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(431, name=name, description=description)


class UnavailableForLegalReasons(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        if name is None:
            name = "Unavailable For Legal Reasons"  # For py < 3.8
        if description is None:
            description = "The server is denying access to the resource as a consequence of a legal demand"
        super().__init__(451, name=name, description=description)


class InternalServerError(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(500, name=name, description=description)


class NotImplemented(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(501, name=name, description=description)


class BadGateway(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(502, name=name, description=description)


class ServiceUnavailable(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(503, name=name, description=description)


class GatewayTimeout(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(504, name=name, description=description)


class HTTPVersionNotSupported(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(505, name=name, description=description)


class VariantAlsoNegotiates(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(506, name=name, description=description)


class InsufficientStorage(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(507, name=name, description=description)


class LoopDetected(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(508, name=name, description=description)


class NotExtended(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(510, name=name, description=description)


class NetworkAuthenticationRequired(HTTPException):
    def __init__(self, name: str = None, description: str = None):
        super().__init__(511, name=name, description=description)



import inspect
import re
import typing

import cachetools

from .converters import (
    Converter,
    FloatConverter,
    IntegerConverter,
    PathConverter,
    StringConverter,
)
from .httpexceptions import MethodNotAllowed, NotFound
from .types import Handler
from .view import View


class Route:
    PARAM_REGEX = re.compile(
        r"<(?P<name>\w+)(?::(?P<type>\w+)(?:\((?P<args>(?:\w+=(?:\w|\+|-|\.)+,?\s*)*)\))?)?>"  # noqa: E501
    )
    PARAM_ARGS_REGEX = re.compile(r"(\w+)=((?:\w|\+|-|\.)+)")
    PARAM_CONVERTERS = {
        "str": StringConverter,
        "path": PathConverter,
        "int": IntegerConverter,
        "float": FloatConverter,
    }

    def __init__(
        self,
        path: str,
        name: str,
        handler: Handler,
        methods: typing.List[str],
        defaults: typing.Dict[str, typing.Any] = None,
    ):
        self.path = path
        self.name = name
        self.handler = handler
        self.methods = methods
        self.defaults = defaults or {}

        handler_signature = inspect.signature(self.handler)
        self.handler_kwargs = [
            param.name
            for param in handler_signature.parameters.values()
            if param.kind in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY)
        ]
        self.handler_is_class = isinstance(self.handler, View)

        if self.name is None:
            self.name = (
                self.handler.__class__.__name__
                if self.handler_is_class
                else self.handler.__name__
            )

        self.converters = {}  # name: converter
        self.index_converters = {}  # index: (name, converter)
        self.build_converters()

        self.regex = re.compile("")
        self.build_regex()

    def build_converters(self):
        segments = self.path.strip("/").split("/")
        for index, segment in enumerate(segments):
            param = self.PARAM_REGEX.fullmatch(segment)
            if param is None:
                continue

            groups = param.groupdict()
            if groups["type"] is None:
                groups["type"] = "str"

            if groups["type"] not in self.PARAM_CONVERTERS:
                raise ValueError(
                    "Expected type to be one of: {}. Got {}".format(
                        ", ".join(self.PARAM_CONVERTERS), groups["type"]
                    )
                )
            converter: Converter = self.PARAM_CONVERTERS[groups["type"]]

            kwargs = {}
            if "args" in groups and groups["args"] is not None:
                args = self.PARAM_ARGS_REGEX.findall(groups["args"])
                for name, value in args:
                    if value in ["True", "False"]:
                        value = True if value == "True" else False
                    elif value.lstrip("+-").isdecimal():
                        value = int(value)
                    elif value.lstrip("+-").replace(".", "", 1).isdecimal():
                        value = float(value)
                    else:
                        value = value.strip("'\"")
                    kwargs[name] = value

            self.converters[groups["name"]] = converter(**kwargs)
            self.index_converters[index] = (groups["name"], converter(**kwargs))

    def build_regex(self):
        segments = self.path.strip("/").split("/")
        regex = ""

        for index, segment in enumerate(segments):
            regex += r"\/"
            if index in self.index_converters:
                name, converter = self.index_converters[index]
                regex += "(?P<{}>{})".format(name, converter.REGEX)

                if index == len(segments) - 1 and name in self.defaults:
                    regex += "?"

            else:
                regex += re.escape(segment)

        regex += r"\/?"

        self.regex = re.compile(regex)

    def match(self, path: str) -> bool:
        return self.regex.fullmatch(path if path.endswith("/") else path + "/")

    def convert(self, path: str) -> typing.Dict[str, typing.Any]:
        kwargs = self.defaults.copy()
        match = self.regex.fullmatch(path if path.endswith("/") else path + "/")

        if match is None:
            raise ValueError("Path doesn't match router path")

        parameters = match.groupdict()

        for name, value in parameters.items():
            if value is None:
                if name in kwargs:
                    continue

            converter = self.converters[name]
            try:
                kwargs[name] = converter.convert(value)
            except ValueError as conversion_error:
                raise ValueError(
                    f"Failed to convert {name} argument: "
                    + str(conversion_error)
                ) from conversion_error

        return kwargs


class Router:
    def __init__(self, routes: typing.Optional[typing.List[Route]] = None):
        self.routes: typing.List[Route] = routes or []
        self._cache: typing.Mapping[str, Route] = cachetools.LFUCache(256)

    def add_route(
        self,
        handler: Handler,
        path: str,
        methods: typing.List[str] = None,
        name: str = None,
        defaults: dict = None,
    ) -> Route:
        route = Route(
            path=path,
            name=name,
            handler=handler,
            methods=methods,
            defaults=defaults or {},
        )
        self.routes.append(route)
        return route

    def get(self, path: str, method: str) -> Route:
        route = self._cache.get(method + " " + path)
        if route is None:
            for possible_route in self.routes:
                if possible_route.match(path):
                    route = possible_route
                    if method not in route.methods:
                        continue
                    break

            if route is None:
                raise NotFound()

        if method not in route.methods:
            raise MethodNotAllowed()

        self._cache[method + " " + path] = route

        return route



import inspect
import typing

from .httpexceptions import MethodNotAllowed
from .types import Handler


class View:
    METHODS = [
        "GET",
        "HEAD",
        "POST",
        "PUT",
        "DELETE",
        "CONNECT",
        "OPTIONS",
        "TRACE",
        "PATCH",
    ]

    def __init__(self, app):
        self.app = app
        self.methods: typing.List[str] = []
        for method in self.METHODS:
            if hasattr(self, method.lower()):
                self.methods.append(method)

        self.methods_kwargs = {}
        for method in self.methods:
            handler_signature = inspect.signature(getattr(self, method.lower()))
            self.methods_kwargs[method] = [
                param.name
                for param in handler_signature.parameters.values()
                if param.kind
                in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY)
            ]

    async def __call__(self, request, **kwargs):
        return await self.dispatch(request, **kwargs)

    async def dispatch(self, request, **kwargs):
        if not hasattr(self, request.method.lower()):
            raise MethodNotAllowed()
        handler: Handler = getattr(self, request.method.lower())

        kwargs["request"] = request
        kwargs = {
            k: v
            for k, v in kwargs.items()
            if k in self.methods_kwargs[request.method]
        }

        return await handler(**kwargs)



import cgi
import os
import pathlib
import typing

from .httpexceptions import NotFound


def get_encoding_from_content_type(content_type):
    """Returns encodings from given Content-Type Header."""

    if not content_type:
        return None

    content_type, params = cgi.parse_header(content_type)

    if "charset" in params:
        return params["charset"].strip("'\"")

    if "text" in content_type:
        return "ISO-8859-1"


def get_encoding_from_headers(headers):
    """Returns encodings from given HTTP Headers."""

    content_type = headers.get("content-type")

    return get_encoding_from_content_type(content_type)


FilePath = typing.Union[bytes, str, os.PathLike]


def file_path_to_path(*paths: FilePath) -> pathlib.Path:
    """Convert a list of paths into a pathlib.Path."""

    safe_paths: typing.List[typing.Union[str, os.PathLike]] = []
    for path in paths:
        if isinstance(path, bytes):
            safe_paths.append(path.decode())
        else:
            safe_paths.append(path)

    return pathlib.Path(*safe_paths)


def safe_join(directory: FilePath, *paths: FilePath) -> pathlib.Path:
    """Safely join the paths to the known directory to return a full path.

    Raises
    ------
        ~baguette.httpexceptions.NotFound
            If the full path does not share a commonprefix with the directory.
    """
    try:
        safe_path = file_path_to_path(directory).resolve(strict=True)
        full_path = file_path_to_path(directory, *paths).resolve(strict=True)
    except FileNotFoundError:
        raise NotFound()
    try:
        full_path.relative_to(safe_path)
    except ValueError:
        raise NotFound()
    return full_path


def split_on_first(text: str, sep: str):
    point = text.find(sep)
    if point == -1:
        return text, text[:0]  # return bytes or str
    return text[:point], text[point + len(sep) :]  # noqa: E203



import json
import typing
from cgi import parse_header
from urllib.parse import parse_qs

from .forms import Field, Form, MultipartForm, URLEncodedForm
from .headers import Headers
from .httpexceptions import BadRequest
from .types import ASGIApp, JSONType, Receive, Scope
from .utils import get_encoding_from_headers

FORM_CONTENT_TYPE = ["application/x-www-form-urlencoded", "multipart/form-data"]


class Request:
    def __init__(self, app: ASGIApp, scope: Scope, receive: Receive):
        self.app = app
        self._scope = scope
        self._receive = receive

        self.http_version: str = scope["http_version"]
        self.asgi_version: str = scope["asgi"]["version"]

        self.headers: Headers = Headers(*scope["headers"])
        self.method: str = scope["method"].upper()
        self.scheme: str = scope["scheme"]
        self.root_path: str = scope["root_path"]
        self.path: str = scope["path"].rstrip("/") or "/"
        self.querystring: typing.Dict[str, typing.List[str]] = parse_qs(
            scope["query_string"].decode("ascii")
        )

        self.server: typing.Tuple[str, int] = scope["server"]
        self.client: typing.Tuple[str, int] = scope["client"]

        # common headers
        self.content_type: str = parse_header(
            self.headers.get("content-type", "")
        )[0]
        self.encoding: str = get_encoding_from_headers(self.headers) or "utf-8"

    async def raw_body(self) -> bytes:
        # caching
        if hasattr(self, "_raw_body"):
            return self._raw_body

        body = b""
        more_body = True

        while more_body:
            message = await self._receive()
            body += message.get("body", b"")
            more_body = message.get("more_body", False)

        self._raw_body = body
        return self._raw_body

    async def body(self) -> str:
        # caching
        if hasattr(self, "_body"):
            return self._body

        raw_body = await self.raw_body()
        self._body = raw_body.decode(self.encoding)
        return self._body

    async def json(self) -> JSONType:
        # caching
        if hasattr(self, "_json"):
            return self._json

        body = await self.body()
        try:
            self._json = json.loads(body)
        except json.JSONDecodeError:
            raise BadRequest(description="Can't decode body as JSON")
        return self._json

    async def form(self, include_querystring: bool = False) -> Form:
        body = await self.raw_body()
        if self.content_type not in FORM_CONTENT_TYPE:
            raise ValueError(
                "Content-type '{}' isn't one of: {}".format(
                    self.content_type, ", ".join(FORM_CONTENT_TYPE)
                )
            )

        if self.content_type == "application/x-www-form-urlencoded":
            form = URLEncodedForm.parse(body.decode(self.encoding))
        elif self.content_type == "multipart/form-data":
            params = parse_header(self.headers.get("content-type", ""))[1]
            form = MultipartForm.parse(
                body,
                boundary=params["boundary"].encode(self.encoding),
                encoding=self.encoding,
            )

        if include_querystring:
            for name, values in self.querystring:
                if name in form.fields:
                    form.fields[name].values.extend(values)
                else:
                    form.fields[name] = Field(name, values)

        return form



import typing

if typing.TYPE_CHECKING:  # pragma: no cover
    from .headers import Headers
    from .request import Request
    from .responses import Response

Scope = typing.MutableMapping[str, typing.Any]
Message = typing.MutableMapping[str, typing.Any]

Receive = typing.Callable[[], typing.Awaitable[Message]]
Send = typing.Callable[[Message], typing.Awaitable[None]]

ASGIApp = typing.Callable[[Scope, Receive, Send], typing.Awaitable[None]]

HeadersType = typing.Union[
    "Headers",
    typing.Mapping[str, str],
    typing.Sequence[typing.Tuple[str, str]],
]

Result = typing.Union[
    "Response",
    typing.Tuple[
        typing.Any,
        typing.Optional[int],
        typing.Optional[HeadersType],
    ],
]

Handler = typing.Callable[["Request"], typing.Awaitable[Result]]

ParamsType = typing.Union[
    str,
    typing.Mapping[str, typing.Union[str, typing.Sequence[str]]],
    typing.Sequence[typing.Tuple[str, typing.Union[str, typing.Sequence[str]]]],
]
BodyType = typing.Union[str, typing.Iterable[str], typing.AsyncIterator[str]]
JSONType = typing.Union[dict, list, tuple, str, int, float, bool]



import typing
from collections.abc import Mapping, Sequence
from json import dumps
from urllib.parse import urlencode

from .app import Baguette
from .headers import Headers, make_headers
from .request import Request
from .responses import Response
from .types import BodyType, HeadersType, JSONType, ParamsType


class TestClient:
    def __init__(
        self,
        app: Baguette,
        default_headers: typing.Optional[HeadersType] = None,
    ):
        self.app = app
        self.default_headers: Headers = make_headers(default_headers)

        self.default_scope = {
            "type": "http",
            "asgi": {"version": "3.0", "spec_version": "2.1"},
            "http_version": "1.1",
            "server": ("127.0.0.1", 8000),
            "client": ("127.0.0.1", 9000),
            "scheme": "http",
            "root_path": "",
        }

    async def request(
        self,
        method: str,
        path: str,
        *,
        params: typing.Optional[ParamsType] = None,
        body: typing.Optional[BodyType] = None,
        json: typing.Optional[JSONType] = None,
        headers: typing.Optional[HeadersType] = None,
    ) -> Response:
        request = self._prepare_request(
            method=method,
            path=path,
            params=params,
            body=body,
            json=json,
            headers=headers,
        )
        response = await self.app.handle_request(request)
        return response

    async def get(
        self,
        path: str,
        *,
        params: typing.Optional[ParamsType] = None,
        body: typing.Optional[BodyType] = None,
        json: typing.Optional[JSONType] = None,
        headers: typing.Optional[HeadersType] = None,
    ) -> Response:
        return await self.request(
            method="GET",
            path=path,
            params=params,
            body=body,
            json=json,
            headers=headers,
        )

    async def head(
        self,
        path: str,
        *,
        params: typing.Optional[ParamsType] = None,
        body: typing.Optional[BodyType] = None,
        json: typing.Optional[JSONType] = None,
        headers: typing.Optional[HeadersType] = None,
    ) -> Response:
        return await self.request(
            method="HEAD",
            path=path,
            params=params,
            body=body,
            json=json,
            headers=headers,
        )

    async def post(
        self,
        path: str,
        *,
        params: typing.Optional[ParamsType] = None,
        body: typing.Optional[BodyType] = None,
        json: typing.Optional[JSONType] = None,
        headers: typing.Optional[HeadersType] = None,
    ) -> Response:
        return await self.request(
            method="POST",
            path=path,
            params=params,
            body=body,
            json=json,
            headers=headers,
        )

    async def put(
        self,
        path: str,
        *,
        params: typing.Optional[ParamsType] = None,
        body: typing.Optional[BodyType] = None,
        json: typing.Optional[JSONType] = None,
        headers: typing.Optional[HeadersType] = None,
    ) -> Response:
        return await self.request(
            method="PUT",
            path=path,
            params=params,
            body=body,
            json=json,
            headers=headers,
        )

    async def delete(
        self,
        path: str,
        *,
        params: typing.Optional[ParamsType] = None,
        body: typing.Optional[BodyType] = None,
        json: typing.Optional[JSONType] = None,
        headers: typing.Optional[HeadersType] = None,
    ) -> Response:
        return await self.request(
            method="DELETE",
            path=path,
            params=params,
            body=body,
            json=json,
            headers=headers,
        )

    async def connect(
        self,
        path: str,
        *,
        params: typing.Optional[ParamsType] = None,
        body: typing.Optional[BodyType] = None,
        json: typing.Optional[JSONType] = None,
        headers: typing.Optional[HeadersType] = None,
    ) -> Response:
        return await self.request(
            method="CONNECT",
            path=path,
            params=params,
            body=body,
            json=json,
            headers=headers,
        )

    async def options(
        self,
        path: str,
        *,
        params: typing.Optional[ParamsType] = None,
        body: typing.Optional[BodyType] = None,
        json: typing.Optional[JSONType] = None,
        headers: typing.Optional[HeadersType] = None,
    ) -> Response:
        return await self.request(
            method="OPTIONS",
            path=path,
            params=params,
            body=body,
            json=json,
            headers=headers,
        )

    async def trace(
        self,
        path: str,
        *,
        params: typing.Optional[ParamsType] = None,
        body: typing.Optional[BodyType] = None,
        json: typing.Optional[JSONType] = None,
        headers: typing.Optional[HeadersType] = None,
    ) -> Response:
        return await self.request(
            method="TRACE",
            path=path,
            params=params,
            body=body,
            json=json,
            headers=headers,
        )

    async def patch(
        self,
        path: str,
        *,
        params: typing.Optional[ParamsType] = None,
        body: typing.Optional[BodyType] = None,
        json: typing.Optional[JSONType] = None,
        headers: typing.Optional[HeadersType] = None,
    ) -> Response:
        return await self.request(
            method="PATCH",
            path=path,
            params=params,
            body=body,
            json=json,
            headers=headers,
        )

    def _prepare_request(
        self,
        method: str,
        path: str,
        *,
        params: typing.Optional[ParamsType] = None,
        body: typing.Optional[BodyType] = None,
        json: typing.Optional[JSONType] = None,
        headers: typing.Optional[HeadersType] = None,
    ) -> Request:
        headers: Headers = self._prepare_headers(headers)
        querystring: str = self._prepare_querystring(params)
        scope = {
            **self.default_scope,
            **{
                "method": method.upper(),
                "path": path,
                "headers": headers.raw(),
                "query_string": querystring.encode("ascii"),
            },
        }

        request = Request(self.app, scope, None)
        request._body = self._prepare_body(body=body, json=json)

        return request

    def _prepare_headers(
        self, headers: typing.Optional[HeadersType] = None
    ) -> Headers:
        headers = make_headers(headers)
        headers = self.default_headers + headers

        return headers

    def _prepare_querystring(
        self, params: typing.Optional[ParamsType] = None
    ) -> str:
        query = {}

        if params is None or isinstance(params, str):
            return params or ""

        if isinstance(params, Mapping):
            params = list(params.items())

        if isinstance(params, Sequence):
            if any(len(param) != 2 for param in params):
                raise ValueError("Incorrect param type")

            for name, value in params:
                if isinstance(value, str):
                    values = [value]
                elif isinstance(value, Sequence):
                    if not all(isinstance(v, str) for v in value):
                        raise ValueError("Incorrect param type")
                    values = list(value)
                else:
                    raise ValueError("Incorrect param type")

                if name in query:
                    query[name].extend(values)
                else:
                    query[name] = values

        else:
            raise ValueError("Incorrect param type")

        querystring = urlencode(query, doseq=True)

        return querystring

    def _prepare_body(
        self,
        body: typing.Optional[BodyType] = None,
        json: typing.Optional[JSONType] = None,
    ) -> str:
        if body is None:
            if json is None:
                return ""

            body = dumps(json)
        return body



import mimetypes
import typing
from cgi import parse_header
from urllib.parse import parse_qs

from .headers import make_headers
from .utils import split_on_first


class Field:
    def __init__(
        self,
        name: str,
        values: typing.List[str],
    ):
        self.name = name
        self.values = values
        if len(self.values) > 0:
            self.value = self.values[0]
        else:
            self.value = None

        self.is_file = False

    def __str__(self) -> str:
        return self.value


class FileField(Field):
    def __init__(
        self,
        name: str,
        content: bytes,
        filename: typing.Optional[str],
        content_type: typing.Optional[str],
        encoding: str = "utf-8",
    ):
        self.name = name

        self.is_file = True
        self.filename = filename
        self.content = content
        self.content_type = content_type
        self.encoding = encoding

        if self.content_type == "application/octet-stream":
            self.content_type = (
                mimetypes.guess_type(self.filename)[0]
                or "application/octet-stream"
            )

    @property
    def text(self) -> str:
        try:
            return self.content.decode(self.encoding)
        except UnicodeDecodeError:
            return self.content

    def __str__(self) -> str:
        return self.text


class Form:
    def __init__(self, fields: typing.Optional[typing.Dict[str, Field]] = None):
        self.fields: typing.Dict[str, Field] = fields or {}

    def __getitem__(self, name: str) -> Field:
        return self.fields[name]

    @classmethod
    def parse(cls, body: str) -> "Form":
        raise NotImplementedError()


class URLEncodedForm(Form):
    @classmethod
    def parse(cls, body: str) -> "URLEncodedForm":
        raw_fields: typing.Dict[str, typing.List[str]] = parse_qs(body)
        fields: typing.Dict[str, Field] = {}
        for name, values in raw_fields.items():
            fields[name] = Field(name, values)
        return cls(fields)


class MultipartForm(Form):
    @classmethod
    def parse(
        cls, body: bytes, boundary: bytes, encoding: str = "utf-8"
    ) -> "MultipartForm":
        fields: typing.Dict[str, Field] = {}
        for part in body.strip(b"\r\n").split(b"".join((b"--", boundary))):
            part = part.strip(b"\r\n")
            if part in (b"", b"--"):  # ignore start and end parts
                continue

            headers, value = split_on_first(part, b"\r\n\r\n")
            headers = make_headers(headers.decode(encoding))
            kwargs = parse_header(headers["content-disposition"])[1]

            name = kwargs["name"]
            is_file = "filename" in kwargs
            filename = kwargs.get("filename", None)

            if name in fields and not fields[name].is_file:
                fields[name].values.append(value)
            else:
                if is_file:
                    fields[name] = FileField(
                        name,
                        value,
                        filename=filename,
                        content_type=parse_header(headers["content-type"])[0],
                        encoding=encoding,
                    )
                else:
                    fields[name] = Field(
                        name,
                        [value],
                    )
        return cls(fields)



"""
Baguette
========

Asynchronous web framework.
"""

from typing import NamedTuple

VersionInfo = NamedTuple(
    "VersionInfo", major=int, minor=int, micro=int, releaselevel=str, serial=int
)

version_info = VersionInfo(major=0, minor=1, micro=5, releaselevel="", serial=0)

__title__ = "baguette"
__author__ = "takos22"
__version__ = "0.1.5"

__all__ = [
    "Baguette",
    "Headers",
    "make_headers",
    "Response",
    "HTMLResponse",
    "PlainTextResponse",
    "JSONResponse",
    "EmptyResponse",
    "FileResponse",
    "make_response",
    "make_error_response",
    "Request",
    "View",
]

from . import testing
from .app import Baguette
from .headers import Headers, make_headers
from .request import Request
from .responses import (
    EmptyResponse,
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    Response,
    make_error_response,
    make_response,
)
from .view import View



import os
import re

from setuptools import setup


def get_version(package):
    """Return package version as listed in `__version__` in `init.py`."""
    path = os.path.join(package, "__init__.py")
    version = ""
    with open(path, "r", encoding="utf8") as init_py:
        version = re.search(
            r"^__version__\s*=\s*['\"]([^'\"]*)['\"]",
            init_py.read(),
            re.MULTILINE,
        ).group(1)

    if not version:
        raise RuntimeError(f"__version__ is not set in {path}")

    return version


def get_packages(package):
    """Return root package and all sub-packages."""
    return [
        dirpath
        for dirpath, *_ in os.walk(package)
        if os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]


def get_long_description(filename: str = "README.rst"):
    """Return the README."""
    with open(filename, "r", encoding="utf8") as readme:
        long_description = readme.read()
    return long_description


def get_requirements(filename: str = "requirements.txt"):
    """Return the requirements."""
    requirements = []
    with open(filename, "r", encoding="utf8") as requirements_txt:
        requirements = requirements_txt.read().splitlines()
    return requirements


extra_requires = {
    "uvicorn": ["uvicorn[standard]"],
    "dev": get_requirements("dev-requirements.txt"),
}

setup(
    name="baguette",
    version=get_version("baguette"),
    url="https://github.com/takos22/baguette",
    license="MIT",
    description="Asynchronous web framework.",
    long_description=get_long_description(),
    long_description_content_type="text/x-rst",
    author="takos22",
    author_email="takos2210@gmail.com",
    packages=get_packages("baguette"),
    python_requires=">=3.6",
    install_requires=get_requirements(),
    extras_require=extra_requires,
    project_urls={
        "Documentation": "https://baguette.readthedocs.io/",
        "Issue tracker": "https://github.com/takos22/baguette/issues",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



import itertools
import typing
from collections.abc import Mapping, Sequence

from .types import HeadersType


class Headers:
    """Headers implementation for handling :class:`str` or :class:`bytes`
    names and values.

    .. tip::
        Use :func:`make_headers` to easily make a header from a :class:`list` or
        a :class:`dict`.
    """

    def __init__(self, *args, **kwargs):
        self._headers: typing.Dict[str, str] = {}

        for name, value in itertools.chain(args, kwargs.items()):
            if isinstance(name, bytes):
                name = name.decode("ascii")
            if isinstance(value, bytes):
                value = value.decode("ascii")
            self._headers[name.lower().strip()] = value.strip()

    def get(self, name, default=None):
        """Gets a header from its name. If not found, returns ``default``.

        Parameters
        ----------
            name: :class:`str` or :class:`bytes`
                Name of the header to get.
            default: Optional anything
                Value to return if header not found.
                Default: ``None``

        Returns
        -------
            :class:`str`
                Header value if found.
            Anything
                ``default``'s value.
        """

        if isinstance(name, bytes):
            name = name.decode("ascii")
        return self._headers.get(name.lower(), default)

    def keys(self):
        """Returns an iterator over the headers names.

        Returns
        -------
            Iterator of :class:`str`
                Iterator over the headers names.
        """

        return self._headers.keys()

    def items(self):
        """Returns an iterator over the headers names and values.

        Returns
        -------
            Iterator of ``(name, value)`` :class:`tuple`
                Iterator over the headers names and values.
        """

        return self._headers.items()

    def raw(self):
        """Returns the raw headers, a :class:`list` of ``[name, value]`` where
        ``name`` and ``value`` are :class:`bytes`.

        Returns
        -------
            :class:`list` of ``[name, value]`` where ``name`` and ``value`` are\
            :class:`bytes`
                Raw headers for the ASGI response.
        """
        return [
            [name.encode("ascii"), value.encode("ascii")]
            for name, value in self
        ]

    def __str__(self) -> str:
        return "\n".join(name + ": " + value for name, value in self)

    def __iter__(self):
        return iter(self.items())

    def __len__(self):
        return len(self._headers)

    def __getitem__(self, name):
        if isinstance(name, bytes):
            name = name.decode("ascii")
        return self._headers[name.lower()]

    def __setitem__(self, name, value):
        if isinstance(name, bytes):
            name = name.decode("ascii")
        if isinstance(value, bytes):
            value = value.decode("ascii")
        self._headers[name.lower()] = value

    def __delitem__(self, name):
        if isinstance(name, bytes):
            name = name.decode("ascii")
        del self._headers[name.lower()]

    def __contains__(self, name):
        return name in self._headers

    def __add__(self, other: typing.Union["Headers", typing.Mapping[str, str]]):
        new = Headers(**self)
        new += other
        return new

    def __iadd__(
        self, other: typing.Union["Headers", typing.Mapping[str, str]]
    ):
        for name, value in other.items():
            if isinstance(name, bytes):
                name = name.decode("ascii")
            if isinstance(value, bytes):
                value = value.decode("ascii")
            self._headers[name.lower()] = value
        return self


def make_headers(headers: HeadersType = None) -> Headers:
    """Makes a :class:`Headers` object from a :class:`list` of
        ``(str, str)`` tuples, a :class:`dict`, or a :class:`Headers` instance.

        Parameters
        ----------
            headers: :class:`list` of ``(str, str)`` tuples, \
            :class:`dict` or :class:`Headers`
                The raw headers to convert.

        Raises
        ------
            :exc:`ValueError`
                ``type_`` isn't one of: 'plain', 'json', 'html'.

        Returns
        -------
            :class:`Headers`
                The converted headers.
        """

    if headers is None:
        headers = Headers()
    elif isinstance(headers, str):
        headers = Headers(
            *[header.split(":") for header in headers.splitlines()]
        )
    elif isinstance(headers, Sequence):
        headers = Headers(*headers)
    elif isinstance(headers, (Mapping, Headers)):
        headers = Headers(**headers)
    else:
        raise ValueError(
            "headers must be a list, a dict, a Headers instance or None"
        )

    return headers



import json
import mimetypes
import re
import typing
import zlib

import aiofiles

from .headers import Headers, make_headers
from .httpexceptions import HTTPException, NotFound
from .types import Result
from .utils import safe_join

HTML_TAG_REGEX = re.compile(r"<\s*\w+[^>]*>.*?<\s*/\s*\w+\s*>")


class Response:
    CHARSET = "utf-8"

    def __init__(
        self,
        body: typing.Union[str, bytes],
        status_code: int = 200,
        headers: typing.Union[dict, Headers] = None,
    ):
        if isinstance(body, str):
            self.text: str = body
            self.body: bytes = body.encode(self.CHARSET)
        elif isinstance(body, bytes):
            self.text: str = body.decode(self.CHARSET)
            self.body: bytes = body
        else:
            raise ValueError("body must be str or bytes")

        self.status_code = status_code
        self.headers = make_headers(headers)

    async def send(self, send):
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.headers.raw(),
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": self.body,
            }
        )


class JSONResponse(Response):
    def __init__(
        self,
        data: typing.Any,
        status_code: int = 200,
        headers: typing.Union[dict, Headers] = None,
    ):
        self.json = data
        body: str = json.dumps(data)
        headers = make_headers(headers)
        headers["content-type"] = "application/json"
        super().__init__(body, status_code, headers)


class PlainTextResponse(Response):
    def __init__(
        self,
        text: typing.Union[str, bytes],
        status_code: int = 200,
        headers: typing.Union[dict, Headers] = None,
    ):
        headers = make_headers(headers)
        headers["content-type"] = "text/plain; charset=" + self.CHARSET
        super().__init__(text, status_code, headers)


class HTMLResponse(Response):
    def __init__(
        self,
        html: typing.Union[str, bytes],
        status_code: int = 200,
        headers: typing.Union[dict, Headers] = None,
    ):
        headers = make_headers(headers)
        headers["content-type"] = "text/html; charset=" + self.CHARSET
        super().__init__(html, status_code, headers)


class EmptyResponse(PlainTextResponse):
    def __init__(
        self,
        status_code: int = 204,
        headers: typing.Union[dict, Headers] = None,
    ):
        super().__init__("", status_code, headers)


class FileResponse(Response):
    def __init__(
        self,
        *paths,
        attachment_filename: typing.Optional[str] = None,
        mimetype: typing.Optional[str] = None,
        as_attachment: bool = False,
        add_etags: bool = True,
        status_code: int = 200,
        headers: typing.Union[dict, Headers] = None,
    ):
        self.mimetype = mimetype
        self.attachment_filename = attachment_filename
        self.headers = make_headers(headers)
        self.status_code = status_code

        self.file_path = safe_join(*paths)
        if not self.file_path.is_file():
            raise NotFound()

        self.file_size = self.file_path.stat().st_size
        self.headers["content-length"] = str(self.file_size)

        if self.attachment_filename is None:
            self.attachment_filename = self.file_path.name

        if self.mimetype is None and self.attachment_filename is not None:
            self.mimetype = (
                mimetypes.guess_type(self.attachment_filename)[0]
                or "application/octet-stream"
            )

        self.headers["content-type"] = self.mimetype

        if as_attachment:
            self.headers[
                "Content-Disposition"
            ] = f"attachment; filename={attachment_filename}"

        if add_etags:
            etag = "{}-{}-{}".format(
                self.file_path.stat().st_mtime,
                self.file_path.stat().st_size,
                zlib.adler32(bytes(self.file_path)),
            )
            self.headers["etag"] = etag

    async def send(self, send):
        async with aiofiles.open(self.file_path, "rb") as f:
            file_body = await f.read()

        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.headers.raw(),
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": file_body,
            }
        )


def make_response(result: Result) -> Response:
    """Makes a :class:`Response` object from a handler return value.

    Parameters
    ----------
        result: Anything described in :ref:`responses`
            The handler return value.

    Returns
    -------
        :class:`Response`
            The handler response.
    """

    if issubclass(type(result), Response):
        return result

    if isinstance(result, tuple):
        body, status_code_or_headers, headers = result + (None,) * (
            3 - len(result)
        )
    else:
        body = result
        status_code_or_headers = None
        status_code = None
        headers = None

    if status_code_or_headers is None:
        status_code = None
    elif isinstance(status_code_or_headers, int):
        status_code = status_code_or_headers
    elif isinstance(status_code_or_headers, (list, dict, Headers)):
        headers = status_code_or_headers
        status_code = None

    headers = make_headers(headers)

    if isinstance(body, (list, dict)):
        response = JSONResponse(body, status_code or 200, headers)
    elif isinstance(body, str):
        if HTML_TAG_REGEX.search(body) is not None:
            response = HTMLResponse(body, status_code or 200, headers)
        else:
            response = PlainTextResponse(body, status_code or 200, headers)
    elif body is None:
        response = EmptyResponse(status_code or 204, headers)
        # print(
        #     RuntimeWarning(
        #         "The value returned by the view function shouldn't be None,"  # noqa: E501
        #         " but instead an empty string and a 204 status code or an EmptyResponse() instance"  # noqa: E501
        #     )
        # )
    else:
        response = PlainTextResponse(str(body), status_code or 200, headers)

    return response


def make_error_response(
    http_exception: HTTPException,
    type_: str = "plain",
    include_description: bool = True,
    traceback: str = None,
) -> Response:
    """Convert an :class:`~baguette.httpexceptions.HTTPException` to a
    :class:`Response`.

    Arguments
    ---------
        http_exception: :class:`~baguette.httpexceptions.HTTPException`
            The HTTP exception to convert to a response.

        type_: :class:`str`
            Type of response. Must be one of: 'plain', 'json', 'html'.

        include_description: :class:`bool`
            Whether to include the description in the response.

        traceback: Optional[:class:`str`]
            Error traceback, usually only included in debug mode.

    Raises
    ------
        :exc:`ValueError`
            ``type_`` isn't one of: 'plain', 'json', 'html'.

    Returns
    -------
        :class:`Response`
            Response that describes the error.
    """

    if type_ == "plain":
        text = http_exception.name
        if include_description and http_exception.description:
            text += ": " + http_exception.description
        if traceback is not None:
            text += "\n" + traceback
        return PlainTextResponse(text, http_exception.status_code)

    elif type_ == "json":
        data = {
            "error": {
                "status": http_exception.status_code,
                "message": http_exception.name,
            }
        }
        if include_description and http_exception.description:
            data["error"]["description"] = http_exception.description
        if traceback is not None:
            data["error"]["traceback"] = traceback
        return JSONResponse(data, http_exception.status_code)

    elif type_ == "html":
        html = f"<h1>{http_exception.status_code} {http_exception.name}</h1>"
        if include_description and http_exception.description:
            html += f"\n<h2>{http_exception.description}</h2>"
        if traceback is not None:
            html += f"\n<pre><code>{traceback}</code></pre>"
        return HTMLResponse(html, http_exception.status_code)

    else:
        raise ValueError(
            "Bad response type. Must be one of: 'plain', 'json', 'html'"
        )
